#define _CRT_SECURE_NO_WARNINGS
#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <string.h>
#include <math.h>



unsigned char *xm;



int read_word(int off)
{
	return xm[off]+(xm[off+1]<<8);
}



int read_dword(int off)
{
	return xm[off]+(xm[off+1]<<8)+(xm[off+2]<<16)+(xm[off+3]<<24);
}



void cleanup(void)
{
	free(xm);
}



int main(int argc, char* argv[])
{
	const float notes[12]={2093.0f,2217.4f,2349.2f,2489.0f,2637.0f,2793.8f,2960.0f,3136.0f,3322.4f,3520.0f,3729.2f,3951.0f};
	const float cputime=(113.0f+123.0f)/2.0f;
	const int pattern_len=8;
	FILE *file;
	int size;
	int order_len;
	int order_loop;
	int channels;
	int patterns;
	int tempo;
	int bpm;
	int i,j,k,l,pp,patlen;
	int tag,note,ins,vol,fx,param;
	char name[1024];
	unsigned char song[256*256][2];
	unsigned char ptdata[256][pattern_len];
	unsigned char order[256*256][2];
	unsigned char row[2];
	int song_len,match,speed;
	unsigned char freqtable[96];
	float div,step;

	if(argc<2)
	{
		printf("xm2huby converter by Shiru (shiru@mail.ru) 04'11\n");
		printf("Usage: xm2huby filename.xm\n");
		exit(0);
	}

	file=fopen(argv[1],"rb");

	if(!file)
	{
		printf("Error: Can't open file %s\n",argv[1]);
		exit(-1);
	}

	fseek(file,0,SEEK_END);
	size=ftell(file);
	fseek(file,0,SEEK_SET);

	xm=(unsigned char*)malloc(size);
	fread(xm,size,1,file);
	fclose(file);

	atexit(cleanup);

	if(memcmp(xm,"Extended Module: ",17))
	{
		printf("Error: Not XM module\n");
		exit(-1);
	}

	order_len=read_word(60+4);
	order_loop=read_word(60+6);
	channels=read_word(60+8);
	patterns=read_word(60+10);
	tempo=read_word(60+16);
	bpm=read_word(60+18);

	if(!order_len)
	{
		printf("Error: Module should have at least one order position\n");
		exit(-1);
	}

	if(channels<2)
	{
		printf("Error: Module should have at least two channels\n");
		exit(-1);
	}

	div=32;

	for(i=0;i<8;i++)
	{
		for(j=0;j<12;j++)
		{
			step=(3500000.0f/(cputime/2.0f))/(notes[j]/div);
			if(step==1) step=0;
			if(step>255) step=0;
			if(step==0x2c) step+=1;
			freqtable[i*12+j]=(unsigned char)step;
		}
		div/=2;
	}

	speed=(int)(2500.0f*(float)tempo*(3500000.0f/1000.0f/cputime)/(float)bpm);

	if(speed<1||speed>65536)
	{
		printf("Warning: Tempo or BPM is out of range\n");
	}

	song_len=0;
	row[0]=0;
	row[1]=0;

	for(i=0;i<order_len;i++)
	{
		pp=60+20+256;
		for(j=0;j<xm[60+20+i];j++) pp=pp+read_dword(pp)+read_word(pp+7);

		patlen=read_word(pp+5);
		pp+=read_dword(pp);

		if(patlen&(pattern_len-1))
		{
			printf("Warning: Patterns length should be multiple of 8 (ptn:#%2.2x)\n",xm[60+20+i]);
		}

		for(j=0;j<patlen;j++)
		{
			for(k=0;k<channels;k++)
			{
				if(xm[pp]&0x80)
				{
					tag=xm[pp++];
					if(tag&0x01) note =xm[pp++]; else note =0;
					if(tag&0x02) ins  =xm[pp++]; else ins  =0;
					if(tag&0x04) vol  =xm[pp++]; else vol  =0;
					if(tag&0x08) fx   =xm[pp++]; else fx   =0;
					if(tag&0x10) param=xm[pp++]; else param=0;
				}
				else
				{
					note =xm[pp++];
					ins  =xm[pp++];
					vol  =xm[pp++];
					fx   =xm[pp++];
					param=xm[pp++];
				}
				if(k<2)
				{
					if(!note)
					{
						song[song_len][k]=row[k];
					}
					if(note>0&&note<97&&ins==1)
					{
						song[song_len][k]=freqtable[note-1];
						if(song[song_len][k]==0)
						{
							printf("Warning: Note out of range (ptn:%2.2x row:%2.2x chn:%i)\n",i,j,k);
						}

						if(fx==0x0e&&(param&0xf0)==0x50)
						{
							song[song_len][k]+=(param&15)-8;
							if(song[song_len][k]<0||song[song_len][k]>=255)
							{
								if(row[k]<0) row[k]=0;
								if(row[k]>255) row[k]=255;
								if(row[k]==0x2c) row[k]++;
								printf("Warning: Note out of range (ptn:%2.2x row:%2.2x chn:%i)\n",i,j,k);
							}
						}
						row[k]=song[song_len][k];
					}
					if(note==97)
					{
						song[song_len][k]=0;
						row[k]=0;
					}
				}
				if(ins==2)
				{
					switch(note)
					{
					case 49: song[song_len][0]=0x2c; break;//C-4
					}
				}
			}

			song_len++;
		}
	}

	patterns=0;
	order_len=0;

	for(i=0;i<song_len/pattern_len;i++)
	{
		for(l=0;l<2;l++)
		{
			match=-1;

			for(j=0;j<patterns;j++)
			{
				match=j;

				for(k=0;k<pattern_len;k++)
				{
					if(song[i*pattern_len+k][l]!=ptdata[j][k])
					{
						match=-1;
						break;
					}
				}

				if(match>=0) break;
			}

			if(match<0)
			{
				order[order_len][l]=patterns;
				for(k=0;k<pattern_len;k++)
				{
					ptdata[patterns][k]=song[i*pattern_len+k][l];
				}
				patterns++;
			}
			else
			{
				order[order_len][l]=match;
			}
		}

		order_len++;
	}

	if(patterns>255)
	{
		printf("Warning: Too many patterns (%i of 255 max)\n",patterns);
	}

	strcpy(name,argv[1]);
	name[strlen(argv[1])-2]='a';
	name[strlen(argv[1])-1]='s';
	name[strlen(argv[1])  ]='m';
	name[strlen(argv[1])+1]=0;

	file=fopen(name,"wt");

	fprintf(file,"module\n");
	fprintf(file,"\tdw #%4.4x\n",speed);
	fprintf(file,"\tdw .p-8\n");

	for(i=0;i<order_len;i++)
	{
		fprintf(file,"\tdb #%2.2x,#%2.2x\n",order[i][0]+1,order[i][1]+1);
	}

	fprintf(file,"\tdb #00\n");
	fprintf(file,".p\n");

	for(i=0;i<patterns;i++)
	{
		fprintf(file,"\tdb ");
		for(j=0;j<pattern_len;j++) fprintf(file,"#%2.2x%c",ptdata[i][j],j<pattern_len-1?',':'\n');
	}

	fclose(file);

	exit(0);
}